from collections import OrderedDict
import sys

PY3 = sys.version_info > (3,)


def unpack(value):
    """Return a three tuple of data, code, and headers"""
    if not isinstance(value, tuple):
        return value, 200, {}
    value_len = len(value)
    if value_len == 2:
        return value[0], value[1], {}
    elif value_len == 3:
        return value[0], value[1], value[2]
    else:
        return value, 200, {}


def get_accept_mimetypes(request):
    accept_types = request.headers.get('accept', None)
    if accept_types is None:
        return {}
    split_types = str(accept_types).split(',')
    # keep the order they appear!
    return OrderedDict([((s, 1,), s,) for s in split_types])


def best_match_accept_mimetype(request, representations, default=None):
    if representations is None or len(representations) < 1:
        return default
    try:
        accept_mimetypes = get_accept_mimetypes(request)
        if accept_mimetypes is None or len(accept_mimetypes) < 1:
            return default
        # find exact matches, in the order they appear in the `Accept:` header
        for accept_type, qual in accept_mimetypes:
            if accept_type in representations:
                return accept_type
        # match special types, like "application/json;charset=utf8" where the first half matches.
        for accept_type, qual in accept_mimetypes:
            type_part = str(accept_type).split(';', 1)[0]
            if type_part in representations:
                return type_part
        # if _none_ of those don't match, then fallback to wildcard matching
        for accept_type, qual in accept_mimetypes:
            if accept_type == "*"\
               or accept_type == "*/*"\
               or accept_type == "*.*":
                return default
    except (AttributeError, KeyError):
        return default



from collections import OrderedDict, Mapping, Sequence

from sanic.request import Request
from sanic.response import BaseHTTPResponse
from sanic.views import HTTPMethodView
from sanic_restful.util import best_match_accept_mimetype, unpack


class Resource(HTTPMethodView):
    """
    Represents an abstract RESTful resource. Concrete resources should
    extend from this class and expose methods for each supported HTTP
    method. If a resource is invoked with an unsupported HTTP method,
    the API will return a response with status 405 Method Not Allowed.
    Otherwise the appropriate method is called and passed all arguments
    from the url rule used when adding the resource to an Api instance. See
    :meth:`~sanic_restful.Api.add_resource` for details.

    :param method_decorators: Mapping class; if you need use Sequence,
        use decorators attribute.
        example:
            method_decorators = {'get': login_require}
            method_decoratros = {'get': [permission, login_require]}
    """
    representations = None
    method_decorators = {}

    def __init__(self, request: Request, *args, **kwargs):
        self.request = request

    async def dispatch_request(self, request: Request, *args, **kwargs):
        method = request.method.lower()
        handler = getattr(self, method, None)

        # if not handler and request.method == "head":
        #     handler = getattr(self, "get", None)
        # assert handler is not None, 'Unimplemented method %r' % request.method

        if self.method_decorators:
            for decorator in self.method_decorators.get(method, []):
                handler = decorator(handler)

        resp = await handler(request, *args, **kwargs)
        if isinstance(resp, BaseHTTPResponse):
            return resp

        representations = self.representations or OrderedDict()

        mediatype = best_match_accept_mimetype(
            request, representations, default=None)
        if mediatype in representations:
            data, code, headers = unpack(resp)
            resp = representations[mediatype](data, code, headers)
            resp.headers['Content-Type'] = mediatype
        return resp

    @classmethod
    def as_view(cls, *class_args, **class_kwargs):
        """Return view function for use with the routing system, that
        dispatches request to appropriate handler method.
        """
        if cls.method_decorators:
            if not isinstance(cls.method_decorators, Mapping):
                raise TypeError("method_decorators must be Mapping")
            for method, decorators in cls.method_decorators.items():
                if not isinstance(decorators, Sequence):
                    cls.method_decorators[method] = [decorators]

        def view(*args, **kwargs):
            self = view.view_class(*class_args, **class_kwargs)
            return self.dispatch_request(*args, **kwargs)

        if cls.decorators:
            view.__module__ = cls.__module__
            for decorator in cls.decorators:
                view = decorator(view)

        view.view_class = cls
        view.__doc__ = cls.__doc__
        view.__module__ = cls.__module__
        view.__name__ = cls.__name__
        return view



from functools import partial
from json import dumps
json_dumps = partial(dumps, separators=(",", ":"))

from sanic_restful.util import PY3

from sanic.response import HTTPResponse


def output_json(app, data, code, headers=None):
    settings = app.config.get('RESTFUL_JSON', {})
    dumps = settings.pop('JSON_DUMP', None) or json_dumps
    # If we're in debug mode, and the indent is not set, we set it to a
    # reasonable value here.  Note that this won't override any existing value
    # that was set.  We also set the "sort_keys" value.
    if app.debug:
        settings.setdefault('indent', 4)
        settings.setdefault('sort_keys', not PY3)

    return HTTPResponse(
        dumps(data, **settings) + "\n",
        headers=headers,
        status=200,
        content_type="application/json",
    )



from collections import OrderedDict
from functools import wraps

from sanic_restful import Resource
from sanic_restful.util import unpack


def marshal(data, fields, envelope=None):
    """Takes raw data (in the form of a dict, list, object) and a dict of
    fields to output and filters the data based on those fields.

    :param data: the actual object(s) from which the fields are taken from
    :param fields: a dict of whose keys will make up the final serialized
                   response output
    :param envelope: optional key that will be used to envelop the serialized
                     response


    >>> from flask_restful import fields, marshal
    >>> data = { 'a': 100, 'b': 'foo' }
    >>> mfields = { 'a': fields.Raw }

    >>> marshal(data, mfields)
    OrderedDict([('a', 100)])

    >>> marshal(data, mfields, envelope='data')
    OrderedDict([('data', OrderedDict([('a', 100)]))])

    """

    def make(cls):
        if isinstance(cls, type):
            return cls()
        return cls

    if isinstance(data, (list, tuple)):
        return (OrderedDict([(envelope, [marshal(d, fields) for d in data])])
                if envelope else [marshal(d, fields) for d in data])

    items = ((k, marshal(data, v)
              if isinstance(v, dict) else make(v).output(k, data))
             for k, v in fields.items())
    return OrderedDict(
        [(envelope, OrderedDict(items))]) if envelope else OrderedDict(items)


class marshal_with(object):
    """A decorator that apply marshalling to the return values of your methods.

    >>> from flask_restful import fields, marshal_with
    >>> mfields = { 'a': fields.Raw }
    >>> @marshal_with(mfields)
    ... def get():
    ...     return { 'a': 100, 'b': 'foo' }
    ...
    ...
    >>> get()
    OrderedDict([('a', 100)])

    >>> @marshal_with(mfields, envelope='data')
    ... def get():
    ...     return { 'a': 100, 'b': 'foo' }
    ...
    ...
    >>> get()
    OrderedDict([('data', OrderedDict([('a', 100)]))])

    see :meth:`flask_restful.marshal`
    """

    def __init__(self, fields, envelope=None):
        """
        :param fields: a dict of whose keys will make up the final
                       serialized response output
        :param envelope: optional key that will be used to envelop the
                        serialized response
        """
        self.fields = fields
        self.envelope = envelope

    def __call__(self, f):
        @wraps(f)
        async def wrapper(*args, **kwargs):
            _cls = args[0] if args else None
            if isinstance(_cls, Resource):
                pass
            resp = await f(*args, **kwargs)
            if isinstance(resp, tuple):
                data, code, headers = unpack(resp)
                return marshal(data, self.fields, self.envelope), code, headers
            else:
                return marshal(resp, self.fields, self.envelope)

        return wrapper


class marshal_with_field:
    """
    A decorator that formats the return values of your methods
     with a single field.

    >>> from flask_restful import marshal_with_field, fields
    >>> @marshal_with_field(fields.List(fields.Integer))
    ... def get():
    ...     return ['1', 2, 3.0]
    ...
    >>> get()
    [1, 2, 3]

    see :meth:`flask_restful.marshal_with`
    """

    def __init__(self, field):
        """
        :param field: a single field with which to marshal the output.
        """
        if isinstance(field, type):
            self.field = field()
        else:
            self.field = field

    def __call__(self, f):
        @wraps(f)
        async def wrapper(*args, **kwargs):
            resp = await f(*args, **kwargs)
            if isinstance(resp, tuple):
                data, code, headers = unpack(resp)
                return self.field.format(data), code, headers
            return self.field.format(resp)

        return wrapper



from collections import OrderedDict
from functools import wraps

from sanic import Blueprint, Sanic
from sanic.exceptions import ServerError
from sanic.response import BaseHTTPResponse, text
from sanic_restful.exceptions import NotAcceptable
from sanic_restful.json import output_json
from sanic_restful.util import unpack, best_match_accept_mimetype

from werkzeug.http import parse_accept_header

DEFAULT_REPRESENTATIONS = [('application/json', output_json)]


class Api:
    """
    The main entry point for the application.
    You need to initialize it with a Flask Application: ::

    >>> app = Flask(__name__)
    >>> api = restful.Api(app)

    Alternatively, you can use :meth:`init_app` to set the Flask application
    after it has been constructed.

    :param app: the Flask application object
    :type app: flask.Flask or flask.Blueprint
    :param prefix: Prefix all routes with a value, eg v1 or 2010-04-01
    :type prefix: str
    :param default_mediatype: The default media type to return
    :type default_mediatype: str
    :param decorators: Decorators to attach to every resource
    :type decorators: list
    :param catch_all_404s: Use :meth:`handle_error`
        to handle 404 errors throughout your app
    :param serve_challenge_on_401: Whether to serve a challenge response to
        clients on receiving 401. This usually leads to a username/password
        popup in web browers.
    :param url_part_order: A string that controls the order that the pieces
        of the url are concatenated when the full url is constructed.  'b'
        is the blueprint (or blueprint registration) prefix, 'a' is the api
        prefix, and 'e' is the path component the endpoint is added with
    :type catch_all_404s: bool
    :param errors: A dictionary to define a custom response for each
        exception or error raised during a request
    :type errors: dict

    """

    def __init__(self,
                 app=None,
                 prefix='',
                 default_mediatype="application/json",
                 decorators=None,
                 catch_all_404s=False,
                 serve_challenge_on_401=False,
                 url_part_order="bae",
                 errors=None):
        self.representations = OrderedDict(DEFAULT_REPRESENTATIONS)
        self.urls = {}
        self.prefix = prefix
        self.default_mediatype = default_mediatype
        self.decorators = decorators if decorators else []
        self.catch_all_404s = catch_all_404s
        self.serve_challenge_on_401 = serve_challenge_on_401
        self.url_part_order = url_part_order
        self.errors = errors or {}
        self.blueprint_setup = None
        self.endpoints = set()
        self.resources = []
        self.app = None
        self.blueprint = None

        if app:
            self.app = app
            self.init_app(app)

    def init_app(self, app):
        """Initialize this class with the given :class:`flask.Flask`
        application or :class:`flask.Blueprint` object.

        :param app: the Flask application or blueprint object

        Examples::

            api = Api()
            api.add_resource(...)
            api.init_app(app)

        """
        # If app is a blueprint, defer the initialization
        if isinstance(app, Blueprint):
            # self.blueprint = app
            self._bp_register = app.register
            # TODO: register api resource for bp that call add resource function
            app.register = self._sanic_blueprint_register_hook(app)
        elif isinstance(app, Sanic):
            self.register_api(app)
        else:
            raise TypeError("only support sanic object and blupirint")

    def _sanic_blueprint_register_hook(self, bp: Blueprint):
        def register(app, options):
            bp_obj = self._bp_register(app, options)
            self.register_api(bp)
            return bp_obj
        return register

    def register_api(self, app):
        if len(self.resources) > 0:
            for resource, urls, kwargs in self.resources:
                self._register_view(app, resource, *urls, **kwargs)

    def _register_view(self, app, resource, *urls, **kwargs):
        endpoint = kwargs.pop("endpoint", None) or resource.__name__.lower()
        self.endpoints.add(endpoint)
        resource_class_args = kwargs.pop("resource_class_args", ())
        resource_class_kwargs = kwargs.pop("resource_class_kwargs", {})

        # Why?
        # resouce.mediatypes = self.mediatypes
        resource.endpoint = endpoint
        resource_func = self.output(
            resource.as_view(self, *resource_class_args,
                             **resource_class_kwargs))

        for decorator in self.decorators:
            resource_func = decorator(resource_func)

        for url in urls:
            rule = self._complete_url(url, '')
            # Add the url to the application or blueprint
            app.add_route(uri=rule, handler=resource_func, **kwargs)

    @property
    def mediatypes(self):
        return [
            "application/json",
            "text/plain; charset=utf-8",
            "application/octet-stream",
            "text/html; charset=utf-8",
        ]

    def output(self, resource):
        """Wraps a resource (as a flask view function), for cases where the
        resource does not directly return a response object

        :param resource: The resource as a flask view function
        """
        @wraps(resource)
        async def wrapper(request, *args, **kwargs):
            resp = await resource(request, *args, **kwargs)
            if isinstance(resp, BaseHTTPResponse):
                return resp
            else:
                data, code, headers = unpack(resp)
            return self.make_response(request, data, code, headers=headers)
        return wrapper

    def make_response(self, request, data, *args, **kwargs):
        """Looks up the representation transformer for the requested media
        type, invoking the transformer to create a response object. This
        defaults to default_mediatype if no transformer is found for the
        requested mediatype. If default_mediatype is None, a 406 Not
        Acceptable response will be sent as per RFC 2616 section 14.1

        :param data: Python object containing response data to be transformed
        """
        default_mediatype = kwargs.pop("fallback_mediatype",
                                       None) or self.default_mediatype
        # mediatype = best_match_accept_mimetype(
        #     request, self.representations, default=default_mediatype)
        mediatype = parse_accept_header(request.headers.get(
            'accept', None)).best_match(
                self.representations, default=default_mediatype)
        if not mediatype:
            raise NotAcceptable("Not Acceptable")
        if mediatype in self.representations:
            resp = self.representations[mediatype](request.app, data, *args,
                                                   **kwargs)
            resp.headers["Content-type"] = mediatype
            return resp
        elif mediatype == "text/plain":
            resp = text(str(data), *args, **kwargs)
            return resp
        else:
            raise ServerError(None)

    def _complete_url(self, url_part, registration_prefix):
        """This method is used to defer the construction of the final url in
        the case that the Api is created with a Blueprint.

        :param url_part: The part of the url the endpoint is registered with
        :param registration_prefix: The part of the url contributed by the
            blueprint.  Generally speaking, BlueprintSetupState.url_prefix
        """
        parts = {'b': registration_prefix, 'a': self.prefix, 'e': url_part}
        return ''.join(parts[key] for key in self.url_part_order if parts[key])

    def add_resource(self, resource, *urls, **kwargs):
        """Adds a resource to the api.

        :param resource: the class name of your resource
        :type resource: :class:`Resource`

        :param urls: one or more url routes to match for the resource, standard
                     flask routing rules apply.  Any url variables will be
                     passed to the resource method as args.
        :type urls: str

        :param endpoint: endpoint name
            (defaults to :meth:`Resource.__name__.lower`
            Can be used to reference this route in :class:`fields.Url` fields
        :type endpoint: str

        :param resource_class_args: args to be forwarded to the constructor of
            the resource.
        :type resource_class_args: tuple

        :param resource_class_kwargs: kwargs to be forwarded to the constructor
            of the resource.
        :type resource_class_kwargs: dict

        Additional keyword arguments not specified above will be passed as-is
        to :meth:`flask.Flask.add_url_rule`.

        Examples::

            api.add_resource(HelloWorld, '/', '/hello')
            api.add_resource(Foo, '/foo', endpoint="foo")
            api.add_resource(FooSpecial, '/special/foo', endpoint="foo")

        """
        if self.app:
            self._register_view(self.app, resource, *urls, **kwargs)
        else:
            self.resources.append((resource, urls, kwargs))

    def resource(self, *urls, **kwargs):
        """Wraps a :class:`~flask_restful.Resource` class, adding it to the
        api. Parameters are the same as :meth:`~flask_restful.Api.add_resource`

        Example::

            app = Flask(__name__)
            api = restful.Api(app)

            @api.resource('/foo')
            class Foo(Resource):
                def get(self):
                    return 'Hello, World!'

        """

        def decorator(cls):
            self.add_resource(cls, *urls, **kwargs)
            return cls

        return decorator

    def representation(self, mediatype):
        """Allows additional representation transformers to be declared for the
        api. Transformers are functions that must be decorated with this
        method, passing the mediatype the transformer represents. Three
        arguments are passed to the transformer:

        * The data to be represented in the response body
        * The http status code
        * A dictionary of headers

        The transformer should convert the data appropriately for the mediatype
        and return a Flask response object.

        Ex::

            @api.representation('application/xml')
            def xml(data, code, headers):
                resp = make_response(convert_data_to_xml(data), code)
                resp.headers.extend(headers)
                return resp
        """

        def wrapper(func):
            self.representations[mediatype] = func
            return func

        return wrapper



from datetime import datetime

import pytest
from sanic import Sanic
from sanic_restful import marshal, fields, Api, Resource, marshal_with
from sanic_restful.marshal import marshal_with_field


address = {
    'country': fields.String,
    'city': fields.String,
}

resource_fields = {
    'name': fields.String,
    'address': {
        'line 1': fields.String(attribute='addr1'),
        'line 2': fields.String(attribute='addr2'),
    },
    'address_region': fields.Nested(address), 
    'first_names': fields.List(fields.String),
    'date_updated': fields.DateTime(dt_format='rfc822'),
    'date_created': fields.DateTime(dt_format='iso8601'),
    'id': fields.Integer,
    'boolean': fields.Boolean,
    'greeting': fields.FormattedString('Hello {name}'),
    'float': fields.Float,
    'arbitrary': fields.Arbitrary,
    'fixed': fields.Fixed,
    'none_int': fields.Integer,
}


@pytest.fixture(scope="module")
def app():
    sanic_app = Sanic(__name__)
    api = Api(sanic_app)

    class TestResource(Resource):
        @marshal_with_field(fields.List(fields.Integer))
        async def get(self, request):
            return ['1', 2, 3.0], 200, {}
        
        @marshal_with_field(fields.List(fields.Integer))
        async def delete(self, request):
            return ['1', 2, 3.0]

        @marshal_with(resource_fields)
        async def post(self, request):
            data = {
                'name': 'bot',
                'line1': 'fake street',
                'line2': 'fake block',
                'city': 1,
                'country': 'China',
                'first_names': ['Emile', 'Raoul'],
                'date_updated': datetime(2019, 1, 1),
                'date_created': datetime(2018, 1, 1),
                'id': '01',
                'boolean': False,
                'float': 1,
                'arbitrary': 634271127864378216478362784632784678324.23432,
                'fixed': 23432,
                'none_int': None
            }
            return data
        
        @marshal_with(resource_fields)
        async def patch(self, request):
            data = {
                'name': 'bot',
                'line1': 'fake street',
                'line2': 'fake block',
                'city': 1,
                'country': 'China',
                'first_names': ['Emile', 'Raoul'],
                'date_updated': datetime(2019, 1, 1),
                'date_created': datetime(2018, 1, 1),
                'id': '01',
                'boolean': False,
                'float': 1,
                'arbitrary': 634271127864378216478362784632784678324.23432,
                'fixed': 23432,
                'none_int': None
            }
            return data, 200, {}

    api.add_resource(TestResource, '/')
    yield sanic_app


class TestMarshal:
    def test_marshal(self, app):
        data = {
            'name': 'bot',
            'addr1': 'fake street',
            'addr2': 'fake block',
            'city_code': 1,
            'first_names': ['Emile', 'Raoul'],
            'date_updated': datetime(2019, 1, 1),
            'date_created': datetime(2018, 1, 1),
            'id': '01',
            'boolean': False,
            'float': 1
        }
        marshal(data, resource_fields)

        request, response = app.test_client.get('/')
        assert response.status == 200

        request, response = app.test_client.post('/')
        assert response.status == 200

        request, response = app.test_client.patch('/')
        assert response.status == 200

        request, response = app.test_client.delete('/')
        assert response.status == 200



import pytest
from sanic import Sanic
from sanic_restful import Api, Resource
from sanic_restful.reqparse import RequestParser


simple_parser = RequestParser()
simple_parser.add_argument('args', location='args')
simple_parser.add_argument('form', location='form')
simple_parser.add_argument('json', location='json')

no_location_parser = RequestParser()
no_location_parser.add_argument('key')

convert_parser = RequestParser()
convert_parser.add_argument('key', type=int, location='args')
convert_parser.add_argument('key1', location='args', type=str)
convert_parser.add_argument('key2', type=str, trim=True, case_sensitive=True)
convert_parser.add_argument(
    'key3', type=str, default='default_value', dest='default_key')


@pytest.fixture(scope="module")
def app():
    sanic_app = Sanic(__name__)
    api = Api(sanic_app)

    class SimpleParserResource(Resource):
        async def post(self, request):
            args = simple_parser.parse_args(request)
            return {'args': args.args, 'form': args.form}

        async def patch(self, request):
            args = simple_parser.parse_args(request)
            return {'json': args.json}

    class NoLocation(Resource):
        async def post(self, request):
            args = no_location_parser.parse_args(request)
            assert args.key
            return {"key": args.key}, 200, {}

    class Convert(Resource):
        async def get(self, request):
            args = convert_parser.parse_args(request)
            assert isinstance(args.key, int)
            assert not args.key1
            assert args.key2 == 'VALUE'
            assert args.default_key == 'default_value'
            return '', 200

        async def post(self, request):
            parser = RequestParser()
            parser.add_argument('key', required=True)
            parser.parse_args(request)
            return '',

    class ParserChoices(Resource):
        async def get(self, request):
            choices = ('one', 'two')
            parser = RequestParser()
            parser.add_argument(
                'foo', choices=choices, help="bad choice: {error_msg}")
            args = parser.parse_args(request)
            assert args.foo in choices
            return ''

    class StrictParse(Resource):
        async def get(self, request):
            parser = RequestParser()
            parser.add_argument('key')
            parser.parse_args(request, True)
            return ''

    class TrimParse(Resource):
        async def get(self, request):
            parser = RequestParser(trim=True)
            parser.add_argument('key')
            args = parser.parse_args(request)
            assert args.key == 'value'
            return ''

    class NullParse(Resource):
        async def post(self, request):
            parser = RequestParser()
            parser.add_argument('key', nullable=False, location='json')
            parser.parse_args(request)
            return ''

        async def put(self, request):
            parser = RequestParser()
            parser.add_argument('key', nullable=True, location='json')
            parser.parse_args(request)
            return ''

    api.add_resource(NullParse, '/null')
    api.add_resource(TrimParse, '/trim')
    api.add_resource(StrictParse, '/strict')
    api.add_resource(ParserChoices, '/choices')
    api.add_resource(Convert, '/convert')
    api.add_resource(NoLocation, '/no_location')
    api.add_resource(SimpleParserResource, '/')
    yield sanic_app


class TestParse:
    def test_parse(self, app: Sanic):
        request, response = app.test_client.post(
            '/',
            params={'args': 'args'},
            data={"form": "form"})
        assert response.status == 200
        assert response.json == {
            'args': "args",
            "form": "form"
        }

    def test_no_location(self, app: Sanic):
        request, response = app.test_client.post(
            '/no_location', json={'key': 'values'})
        assert response.status == 200
        assert response.json == {
            'key': 'values'
        }

    def test_convert(self, app: Sanic):
        request, response = app.test_client.get(
            '/convert', params={'key': '1', 'key2': 'VALUE'})
        assert response.status == 200

        request, response = app.test_client.post(
            '/convert')
        assert response.status == 400

    def test_parser_action(self):
        parser = RequestParser()
        parser.add_argument('key', default='value', dest='key0')
        parser.copy()
        parser.replace_argument('key', default='value1', dest='key0')
        parser.remove_argument('key')

    def test_parser_choice(self, app):
        request, response = app.test_client.get(
            '/choices', params={'foo': 'one'})
        assert response.status == 200
        request, response = app.test_client.get(
            '/choices', params={'foo': 'zero'})
        assert response.status == 400

    def test_parser_strict(self, app):
        request, response = app.test_client.get(
            '/strict', params={
                'key': 'value',
                'other_key': 'other_value'
            })
        assert response.status == 400

    def test_parser_trim(self, app):
        request, response = app.test_client.get(
            '/trim', params={'key': '  value  '})
        assert response.status == 200

    def test_null_parser(self, app):
        requst, response = app.test_client.post(
            '/null', json={'key': None})
        assert response.status == 400
        request, response = app.test_client.put('/null', json={'key': None})
        assert response.status == 200



import pytest
from sanic import Sanic
from sanic.response import json
from sanic_restful import Api, Resource


@pytest.fixture(scope="module")
def app():
    sanic_app = Sanic(__name__)
    api = Api(sanic_app)

    class TestResource(Resource):
        async def get(self, request):
            return {'response': 'get'}

        async def post(self, request):
            return {'response': 'post'}

        async def put(self, request):
            return {'response': 'put'}

        async def delete(self, request):
            return {'response': 'delete'}

        async def patch(self, request):
            return {'response': 'patch'}

    class TestResponse(Resource):
        async def get(self, request):
            return json({'response': 'get'})

    api.add_resource(TestResponse, '/response')
    api.add_resource(TestResource, '/')
    yield sanic_app


class TestAPI:

    def test_api(self, app: Sanic):
        request, response = app.test_client.get('/', debug=True)
        assert response.status == 200
        assert response.json['response'] == 'get'

        request, response = app.test_client.post('/')
        assert response.status == 200
        assert response.json['response'] == 'post'

        request, response = app.test_client.put('/')
        assert response.status == 200
        assert response.json['response'] == 'put'

        request, response = app.test_client.delete('/')
        assert response.status == 200
        assert response.json['response'] == 'delete'

        request, response = app.test_client.patch('/')
        assert response.status == 200
        assert response.json['response'] == 'patch'

        request, response = app.test_client.head('/')
        assert response.status == 405

        request, response = app.test_client.get('/response')
        assert response.status == 200
        assert response.json['response'] == 'get'



from sanic.exceptions import SanicException, add_status_code


@add_status_code(405)
class MethodNotAllowed(SanicException):
    pass


@add_status_code(406)
class NotAcceptable(SanicException):
    pass



from sanic import Sanic
from sanic.exceptions import NotFound, Unauthorized
from sanic.log import error_logger
from sanic_restful import reqparse, Api, Resource

app = Sanic(__name__)
errors = {
    'TypeError': {
        'status': 404,
        'message': 'not found'
    }
}
api = Api(app, errors=errors)


TODOS = {
    'todo1': {'task': 'build an API'},
    'todo2': {'task': '?????'},
    'todo3': {'task': 'profit!'},
}


def abort_if_todo_doesnt_exist(todo_id):
    if todo_id not in TODOS:
        raise NotFound(message="Todo {} doesn't exist".format(todo_id))


parser = reqparse.RequestParser()
parser.add_argument('task', location='form')
parser.add_argument('test', location='json')

post_parser = reqparse.RequestParser()
post_parser.add_argument('key')
# post_parser.add_argument("value")

# Todo
#   show a single todo item and lets you delete them
class Todo(Resource):
    async def get(self, request, todo_id):
        raise NotFound('hahaha')
        abort_if_todo_doesnt_exist(todo_id)
        return TODOS[todo_id]

    async def delete(self, request, todo_id):
        abort_if_todo_doesnt_exist(todo_id)
        del TODOS[todo_id]
        return '', 204

    async def put(self, request, todo_id):
        args = parser.parse_args(request)
        task = {'task': args['task']}
        TODOS[todo_id] = task
        return task, 201

    async def patch(self, request, todo_id):
        raise Unauthorized('fail')


# TodoList
#   shows a list of all todos, and lets you POST to add new tasks
class TodoList(Resource):
    async def get(self, request):
        return TODOS

    async def post(self, request):
        args = parser.parse_args(request)
        error_logger.info(type(request.json.get('test')))
        todo_id = 'todo%d' % (len(TODOS) + 1)
        TODOS[todo_id] = {'task': args['task']}
        return TODOS[todo_id], 201


class TodoArgs(Resource):
    async def post(self, request):
        args = post_parser.parse_args(request)
        TODOS[args.key] = args.value
        return TODOS


# Actually setup the Api resource routing here
api.add_resource(TodoList, '/todos')
api.add_resource(Todo, '/todos/<todo_id:string>')
api.add_resource(TodoArgs, '/todos/args')


if __name__ == '__main__':
    app.run(debug=True)



from calendar import timegm
from decimal import Decimal as MyDecimal, ROUND_HALF_EVEN
from email.utils import formatdate

from sanic_restful.marshal import marshal

__all__ = ["String", "FormattedString", "DateTime", "Float",
           "Integer", "Arbitrary", "Nested", "List", "Raw", "Boolean",
           "Fixed", "Price"]


class MarshallingException(Exception):
    """
    This is an encapsulating Exception in case of marshalling error.
    """

    def __init__(self, underlying_exception):
        # just put the contextual representation of the error to hint on what
        # went wrong without exposing internals
        super().__init__(str(underlying_exception))


def is_indexable_but_not_string(obj):
    return not hasattr(obj, "strip") and hasattr(obj, "__iter__")


def get_value(key, obj, default=None):
    """Helper for pulling a keyed value off various types of objects"""
    if isinstance(key, int):
        return _get_value_for_key(key, obj, default)
    elif callable(key):
        return key(obj)
    else:
        return _get_value_for_keys(key.split('.'), obj, default)


def _get_value_for_keys(keys, obj, default):
    if len(keys) == 1:
        return _get_value_for_key(keys[0], obj, default)
    else:
        return _get_value_for_keys(
            keys[1:], _get_value_for_key(keys[0], obj, default), default)


def _get_value_for_key(key, obj, default):
    if is_indexable_but_not_string(obj):
        try:
            return obj[key]
        except (IndexError, TypeError, KeyError):
            pass
    return getattr(obj, key, default)


def to_marshallable_type(obj):
    """Helper for converting an object to a dictionary only if it is not
    dictionary already or an indexable object nor a simple type"""
    if obj is None:
        return None  # make it idempotent for None

    if hasattr(obj, '__marshallable__'):
        return obj.__marshallable__()

    if hasattr(obj, '__getitem__'):
        return obj  # it is indexable it is ok

    return dict(obj.__dict__)


class Raw(object):
    """Raw provides a base field class from which others should extend. It
    applies no formatting by default, and should only be used in cases where
    data does not need to be formatted before being serialized. Fields should
    throw a :class:`MarshallingException` in case of parsing problem.

    :param default: The default value for the field, if no value is
        specified.
    :param attribute: If the public facing value differs from the internal
        value, use this to retrieve a different attribute from the response
        than the publicly named value.
    """

    def __init__(self, default=None, attribute=None):
        self.attribute = attribute
        self.default = default

    def format(self, value):
        """Formats a field's value. No-op by default - field classes that
        modify how the value of existing object keys should be presented should
        override this and apply the appropriate formatting.

        :param value: The value to format
        :exception MarshallingException: In case of formatting problem

        Ex::

            class TitleCase(Raw):
                def format(self, value):
                    return unicode(value).title()
        """
        return value

    def output(self, key, obj):
        """Pulls the value for the given key from the object, applies the
        field's formatting and returns the result. If the key is not found
        in the object, returns the default value. Field classes that create
        values which do not require the existence of the key in the object
        should override this and return the desired value.

        :exception MarshallingException: In case of formatting problem
        """

        value = get_value(key
                          if self.attribute is None else self.attribute, obj)

        if value is None:
            return self.default

        return self.format(value)


class Nested(Raw):
    """Allows you to nest one set of fields inside another.
    See :ref:`nested-field` for more information

    :param dict nested: The dictionary to nest
    :param bool allow_null: Whether to return None instead of a dictionary
        with null keys, if a nested dictionary has all-null keys
    :param kwargs: If ``default`` keyword argument is present, a nested
        dictionary will be marshaled as its value if nested dictionary is
        all-null keys (e.g. lets you return an empty JSON object instead of
        null)
    """

    def __init__(self, nested, allow_null=False, **kwargs):
        self.nested = nested
        self.allow_null = allow_null
        super().__init__(**kwargs)

    def output(self, key, obj):
        value = get_value(key
                          if self.attribute is None else self.attribute, obj)
        if value is None:
            if self.allow_null:
                return None
            elif self.default is not None:
                return self.default

        return marshal(value, self.nested)


class List(Raw):
    """
    Field for marshalling lists of other fields.

    See :ref:`list-field` for more information.

    :param cls_or_instance: The field type the list will contain.
    """

    def __init__(self, cls_or_instance, **kwargs):
        super().__init__(**kwargs)
        error_msg = ("The type of the list elements must be a subclass of "
                     "flask_restful.fields.Raw")
        if isinstance(cls_or_instance, type):
            if not issubclass(cls_or_instance, Raw):
                raise MarshallingException(error_msg)
            self.container = cls_or_instance()
        else:
            if not isinstance(cls_or_instance, Raw):
                raise MarshallingException(error_msg)
            self.container = cls_or_instance

    def format(self, value):
        # Convert all instances in typed list to container type
        if isinstance(value, set):
            value = list(value)

        return [
            self.container.output(
                idx, val if (isinstance(val, dict) or
                             (self.container.attribute
                              and hasattr(val, self.container.attribute)))
                and not isinstance(self.container, Nested)
                and not type(self.container) is Raw else value)
            for idx, val in enumerate(value)
        ]

    def output(self, key, data):
        value = get_value(key
                          if self.attribute is None else self.attribute, data)
        # we cannot really test for external dict behavior
        if is_indexable_but_not_string(value) and not isinstance(value, dict):
            return self.format(value)

        if value is None:
            return self.default

        return [marshal(value, self.container.nested)]


class String(Raw):
    """
    Marshal a value as a string. Uses ``six.text_type`` so values will
    be converted to :class:`unicode` in python2 and :class:`str` in
    python3.
    """
    def format(self, value):
        try:
            return str(value)
        except ValueError as ve:
            raise MarshallingException(ve)


class Integer(Raw):
    """ Field for outputting an integer value.

    :param int default: The default value for the field, if no value is
        specified.
    """
    def __init__(self, default=0, **kwargs):
        super().__init__(default=default, **kwargs)

    def format(self, value):
        try:
            if value is None:
                return self.default
            return int(value)
        except ValueError as ve:
            raise MarshallingException(ve)


class Boolean(Raw):
    """
    Field for outputting a boolean value.

    Empty collections such as ``""``, ``{}``, ``[]``, etc. will be converted to
    ``False``.
    """
    def format(self, value):
        return bool(value)


class FormattedString(Raw):
    """
    FormattedString is used to interpolate other values from
    the response into this field. The syntax for the source string is
    the same as the string :meth:`~str.format` method from the python
    stdlib.

    Ex::

        fields = {
            'name': fields.String,
            'greeting': fields.FormattedString("Hello {name}")
        }
        data = {
            'name': 'Doug',
        }
        marshal(data, fields)
    """
    def __init__(self, src_str):
        """
        :param string src_str: the string to format with the other
        values from the response.
        """
        super().__init__()
        self.src_str = str(src_str)

    def output(self, key, obj):
        try:
            data = to_marshallable_type(obj)
            return self.src_str.format(**data)
        except (TypeError, IndexError) as error:
            raise MarshallingException(error)


class Float(Raw):
    """
    A double as IEEE-754 double precision.
    ex : 3.141592653589793 3.1415926535897933e-06 3.141592653589793e+24 nan inf
    -inf
    """

    def format(self, value):
        try:
            return float(value)
        except ValueError as ve:
            raise MarshallingException(ve)


class Arbitrary(Raw):
    """
        A floating point number with an arbitrary precision
          ex: 634271127864378216478362784632784678324.23432
    """

    def format(self, value):
        return str(MyDecimal(value))


class DateTime(Raw):
    """
    Return a formatted datetime string in UTC. Supported formats are RFC 822
    and ISO 8601.

    See :func:`email.utils.formatdate` for more info on the RFC 822 format.

    See :meth:`datetime.datetime.isoformat` for more info on the ISO 8601
    format.

    :param dt_format: ``'rfc822'`` or ``'iso8601'``
    :type dt_format: str
    """
    def __init__(self, dt_format='rfc822', **kwargs):
        super().__init__(**kwargs)
        self.dt_format = dt_format

    def format(self, value):
        try:
            if self.dt_format == 'rfc822':
                return _rfc822(value)
            elif self.dt_format == 'iso8601':
                return _iso8601(value)
            else:
                raise MarshallingException(
                    'Unsupported date format %s' % self.dt_format
                )
        except AttributeError as ae:
            raise MarshallingException(ae)


ZERO = MyDecimal()


class Fixed(Raw):
    """
    A decimal number with a fixed precision.
    """
    def __init__(self, decimals=5, **kwargs):
        super().__init__(**kwargs)
        self.precision = MyDecimal('0.' + '0' * (decimals - 1) + '1')

    def format(self, value):
        dvalue = MyDecimal(value)
        if not dvalue.is_normal() and dvalue != ZERO:
            raise MarshallingException('Invalid Fixed precision number.')
        return str(
            dvalue.quantize(self.precision, rounding=ROUND_HALF_EVEN))


"""Alias for :class:`~fields.Fixed`"""
Price = Fixed


def _rfc822(dt):
    """Turn a datetime object into a formatted date.

    Example::
    fields._rfc822(datetime(2011, 1, 1)) => "Sat, 01 Jan 2011 00:00:00 -0000"

    :param dt: The datetime to transform
    :type dt: datetime
    :return: A RFC 822 formatted date string
    """
    return formatdate(timegm(dt.utctimetuple()))


def _iso8601(dt):
    """Turn a datetime object into an ISO8601 formatted date.

    Example::

        fields._iso8601(datetime(2012, 1, 1, 0, 0)) => "2012-01-01T00:00:00"

    :param dt: The datetime to transform
    :type dt: datetime
    :return: A ISO 8601 formatted date string
    """
    return dt.isoformat()



import collections
from copy import deepcopy
import decimal

from sanic.exceptions import abort, InvalidUsage
from sanic.request import Request
from werkzeug.datastructures import MultiDict as RequestParameters


# class RequestParameters(collections.UserDict):
#     """Hosts a dict with lists as values where get returns the first
#     value of the list and getlist returns the whole shebang
#     """
#     def __init__(self, *args, **kwargs):
#         super().__init__(*args, **kwargs)
#         self.data = collections.defaultdict(list)

#     def get(self, name, default=None):
#         """Return the first value, either the default or actual"""
#         return super().get(name, [default])[0]

#     def getlist(self, name, default=None):
#         """Return the entire list"""
#         return super().get(name, default)

#     def __setitem__(self, key, item): self.data[key].append(item)


class Namespace(collections.UserDict):
    def __missing__(self, name):
        raise AttributeError(name)

    def __getattr__(self, name):
        return self.__getitem__(name)


_friendly_location = {
    'json': 'the JSON body',
    'form': 'the post body',
    'args': 'the query string',
    'values': 'the post body or the query string',
    'headers': 'the HTTP headers',
    'cookies': 'the request\'s cookies',
    'files': 'an uploaded file',
}


class Argument(object):

    """
    :param name: Either a name or a list of option strings, e.g. foo or
        -f, --foo.
    :param default: The value produced if the argument is absent from the
        request.
    :param dest: The name of the attribute to be added to the object
        returned by :meth:`~reqparse.RequestParser.parse_args()`.
    :param bool required: Whether or not the argument may be omitted (optionals
        only).
    :param action: The basic type of action to be taken when this argument
        is encountered in the request. Valid options are "store" and "append".
    :param ignore: Whether to ignore cases where the argument fails type
        conversion
    :param type: The type to which the request argument should be
        converted. If a type raises an exception, the message in the
        error will be returned in the response. Defaults to :class:`unicode`
        in python2 and :class:`str` in python3.
    :param location: The attributes of the :class:`flask.Request` object
        to source the arguments from (ex: headers, args, etc.), can be an
        iterator. The last item listed takes precedence in the result set.
    :param choices: A container of the allowable values for the argument.
    :param help: A brief description of the argument, returned in the
        response when the argument is invalid. May optionally contain
        an "{error_msg}" interpolation token, which will be replaced with
        the text of the error raised by the type converter.
    :param bool case_sensitive: Whether argument values in the request are
        case sensitive or not (this will convert all values to lowercase)
    :param bool store_missing: Whether the arguments default value should
        be stored if the argument is missing from the request.
    :param bool trim: If enabled, trims whitespace around the argument.
    :param bool nullable: If enabled, allows null value in argument.
    """

    def __init__(self,
                 name,
                 default=None,
                 dest=None,
                 required=False,
                 ignore=False,
                 type=None,
                 location=('json', 'form', 'args', 'files'),
                 choices=(),
                 action='store',
                 help=None,
                 operators=('=', ),
                 case_sensitive=True,
                 store_missing=True,
                 trim=False,
                 nullable=True,
                 ignore_invalid_usage=True):
        self.name = name
        self.default = default
        self.dest = dest
        self.required = required
        self.ignore = ignore
        self.location = location
        self.type = type
        self.choices = choices
        self.action = action
        self.help = help
        self.case_sensitive = case_sensitive
        self.operators = operators
        self.store_missing = store_missing
        self.trim = trim
        self.nullable = nullable
        self.ignore_invalid_usage = ignore_invalid_usage

    def source(self, request):
        """Pulls values off the request in the provided location
        if location is str:
            json -> dict
            form, args, file -> RequestParameters
        if location is sequence:
            return RequestParameters

        :param request: The flask request object to parse arguments from
        """
        if isinstance(self.location, str):
            try:
                value = getattr(request, self.location, RequestParameters())
            except InvalidUsage as e:
                if self.ignore_invalid_usage:
                    return RequestParameters()
                else:
                    raise e

            if callable(value):
                value = value()
            if value:
                return value
        else:
            values = RequestParameters()
            for l in self.location:
                value = getattr(request, l, None)
                if callable(value):
                    value = value()
                if value:
                    values.update(value)
            return values

        return RequestParameters()

    def convert(self, value, op):
        if self.location == "file":
            return value
        if not value:
            if self.nullable:
                return value
            else:
                raise ValueError("Must not be null")

        try:
            if not self.type:
                return value
            return self.type(value, self.name, op)
        except TypeError:
            try:
                if self.type is decimal.Decimal:
                    return self.type(str(value), self.name)
                else:
                    return self.type(value, self.name)
            except TypeError:
                return self.type(value)

    def handle_validation_error(self, app, error, bundle_errors):
        """Called when an error is raised while parsing. Aborts the request
        with a 400 status and an error message

        :param error: the error that was raised
        :param bundle_errors: do not abort when first error occurs, return a
            dict with the name of the argument and the error message to be
            bundled
        """
        error_msg = self.help.format(error_msg=error) if self.help else error
        msg = {self.name: error_msg}

        if app.config.get("BUNDLE_ERRORS", False) or bundle_errors:
            return error, msg
        abort(400, message=msg)

    def parse(self, request, req_temp, bundle_errors=False):
        """Parses argument value(s) from the request, converting according to
        the argument's type.

        :param request: The flask request object to parse arguments from
        :param do not abort when first error occurs, return a
            dict with the name of the argument and the error message to be
            bundled
        """
        source = self.source(request)

        results = []

        # Sentinels
        _not_found = False
        _found = True

        for operator in self.operators:
            name = self.name + operator.replace("=", "", 1)
            if name in source:
                # Account for MultiDict and regular dict
                if hasattr(source, "getlist"):
                    values = source.getlist(name)
                else:
                    values = source.get(name)
                    if not (isinstance(values, collections.MutableSequence)
                            and self.action == 'append'):
                        values = [values]
                for value in values:
                    if hasattr(value, "strip") and self.trim:
                        value = value.strip()
                    if hasattr(value, "lower") and not self.case_sensitive:
                        value = value.lower()

                        if hasattr(self.choices, "__iter__"):
                            self.choices = [choice.lower()
                                            for choice in self.choices]

                    try:
                        value = self.convert(value, operator)
                    except Exception as error:
                        if self.ignore:
                            continue
                        self.handle_validation_error(
                            request.app, error, bundle_errors)

                    if self.choices and value not in self.choices:
                        if request.app.config.get("BUNDLE_ERRORS",
                                                  False) or bundle_errors:
                            return self.handle_validation_error(
                                request.app,
                                ValueError("{0} is not a valid choice".format(
                                    value)), bundle_errors)
                        self.handle_validation_error(
                            request.app,
                            ValueError(
                                "{0} is not a valid choice".format(value)),
                            bundle_errors)

                    if name in req_temp.unparsed_arguments:
                        req_temp.unparsed_arguments.pop(name)
                    results.append(value)

        if not results and self.required:
            if isinstance(self.location, str):
                error_msg = "Missing required parameter in {0}".format(
                    _friendly_location.get(self.location, self.location)
                )
            else:
                friendly_locations = [_friendly_location.get(loc, loc)
                                      for loc in self.location]
                error_msg = "Missing required parameter in {0}".format(
                    ' or '.join(friendly_locations)
                )
            if request.app.config.get("BUNDLE_ERRORS", False) or bundle_errors:
                return self.handle_validation_error(request.app,
                                                    ValueError(error_msg),
                                                    bundle_errors)
            self.handle_validation_error(request.app, ValueError(error_msg),
                                         bundle_errors)

        if not results:
            if callable(self.default):
                return self.default(), _not_found
            else:
                return self.default, _not_found

        if self.action == 'append':
            return results, _found

        if self.action == 'store' or len(results) == 1:
            return results[0], _found
        return results, _found


class RequestParser:
    """Enables adding and parsing of multiple arguments in the context of a
    single request. Ex::

        from flask_restful import reqparse

        parser = reqparse.RequestParser()
        parser.add_argument('foo')
        parser.add_argument('int_bar', type=int)
        args = parser.parse_args()

    :param bool trim: If enabled, trims whitespace on all arguments in this
        parser
    :param bool bundle_errors: If enabled, do not abort when first error
        occurs, return a dict with the name of the argument and the error
        message to be bundled and return all validation errors
    """

    def __init__(self,
                 argument_cls=Argument,
                 namespace_cls=Namespace,
                 trim=False,
                 bundle_errors=False):
        self.args = {}
        self.argument_cls = argument_cls
        self.namespace_cls = namespace_cls
        self.trim = trim
        self.bundle_errors = bundle_errors

    def add_argument(self, *args, **kwargs) -> None:
        """Adds an argument to be parsed.

        Accepts either a single instance of Argument or arguments to be passed
        into :class:`Argument`'s constructor.

        See :class:`Argument`'s constructor for documentation on the
        available options.
        """
        if len(args) == 1 and isinstance(args[0], self.argument_cls):
            # self.args[args[0].name] = args[0]
            argument_obj = args[0]
        else:
            argument_obj = self.argument_cls(*args, **kwargs)

        if self.trim and self.argument_cls is Argument:
            argument_obj.trim = kwargs.get('trim', self.trim)

        if self.args.get(argument_obj.name):
            raise RuntimeError('Argument is existed')
        else:
            self.args[argument_obj.name] = argument_obj

    def parse_args(self, request: Request, strict=False):
        """Parse all arguments from the provided request and return the results
        as a Namespace

        :param strict: if req includes args not in parser,
                throw 400 BadRequest exception
        """
        namespace = self.namespace_cls()

        # A record of arguments not yet parsed; as each is found
        # among self.args, it will be popped out
        req_temp = collections.namedtuple('RedType', 'unparsed_arguments')
        req_temp.unparsed_arguments = dict(
            self.argument_cls('').source(request)) if strict else {}
        errors = {}

        for name, arg in self.args.items():
            value, found = arg.parse(request, req_temp, self.bundle_errors)
            if isinstance(value, ValueError):
                errors.update(found)
                found = None
            if found or arg.store_missing:
                namespace[arg.dest or name] = value

        if errors:
            abort(400, message=errors)

        if strict and req_temp.unparsed_arguments:
            abort(
                400, 'Unknown arguments: %s' % ', '.join(
                    req_temp.unparsed_arguments.keys()))

        return namespace

    def copy(self):
        """
        Creates a copy of this RequestParser with
        the same set of arguments
        """
        parser_copy = self.__class__(self.argument_cls, self.namespace_cls)
        parser_copy.args = deepcopy(self.args)
        parser_copy.trim = self.trim
        parser_copy.bundle_errors = self.bundle_errors
        return parser_copy

    def replace_argument(self, name, *args, **kwargs):
        """ Replace the argument matching the given name with a new version."""
        new_args = self.argument_cls(name, *args, **kwargs)
        if self.args.get(name):
            self.args[name] = new_args
        else:
            raise AttributeError('%s not existed' % name)

    def remove_argument(self, name):
        self.args.pop(name)



from collections import UserDict, defaultdict


class TypeConversionDict(UserDict):

    def get(self, key, default=None, type=None):
        try:
            rv = self[key]
        except KeyError:
            return default
        if type is not None:
            try:
                rv = type(rv)
            except ValueError:
                rv = default
        return rv


class MultiDict(TypeConversionDict):
    def __init__(self, mapping=None):
        tmp = defaultdict(list)
        for key, value in mapping or ():
            tmp[key].append(value)
        super().__init(self, tmp)



from functools import wraps

import pytest
from sanic import Sanic
from sanic_restful import Api, Resource


@pytest.fixture(scope="module")
def app():
    sanic_app = Sanic(__name__)
    api = Api(sanic_app)

    def login_require(func):
        @wraps(func)
        async def wrapper(request, *args, **kwargs):
            response = await func(request, *args, **kwargs)
            return response
        return wrapper

    class TestApi(Resource):
        method_decorators = {'get': login_require}

        async def get(self, request):
            return ''

    api.add_resource(TestApi, '/')
    yield sanic_app


class TestDecorator:
    def test_decorator(self, app: Sanic):
        request, response = app.test_client.get('/')
        assert response.status == 200



import re
from os import path
from setuptools import setup, find_packages


requirements = [
    'aniso8601>=0.82',
    'Sanic',
    'pytz',
    'werkzeug'
]


version_file = path.join(
    path.dirname(__file__),
    'sanic_restful',
    '__version__.py'
)
with open(version_file, 'r') as fp:
    m = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]",
        fp.read(),
        re.M
    )
    version = m.groups(1)[0]


setup(
    name='sanic-restful',
    version=version,
    license='BSD',
    url='https://github.com/CoCongV/sanic-restful',
    author='CoCong',
    author_email='cong.lv.yx@gmail.com',
    description='''Simple framework for creating REST APIs,'''
    '''Likely Flask-Restful''',
    packages=find_packages(exclude=['tests']),
    classifiers=[
        'Environment :: Web Environment',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: BSD License',
    ],
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    test_suite='nose.collector',
    install_requires=requirements,
    setup_requires=[
        'pytest-runner',
    ],
    tests_require=[
        'sanic-restful', 'pytest', 'pytest_cov', 'nose',
    ],
    # Install these with "pip install -e '.[paging]'" or '.[docs]'
)



# needs: pip install python-simplexml
from simplexml import dumps
from sanic import Sanic
from sanic.response import HTTPResponse
from sanic_restful import Api, Resource


def output_xml(app, data, code, headers=None):
    """Makes a Flask response with a XML encoded body"""
    resp = HTTPResponse(
        dumps({
            'response': data
        }), status=code, headers=headers)
    resp.headers.extend(headers or {})
    return resp


app = Sanic(__name__)
api = Api(app, default_mediatype='application/xml')
api.representations['application/xml'] = output_xml


class Hello(Resource):
    """
        # you need requests
        >>> from requests import get
        >>> get('http://localhost:5000/me').content # default_mediatype
        '<?xml version="1.0" ?><response><hello>me</hello></response>'
        >>> get('http://localhost:5000/me', headers={"accept":"application/json"}).content
        '{"hello": "me"}'
        >>> get('http://localhost:5000/me', headers={"accept":"application/xml"}).content
        '<?xml version="1.0" ?><response><hello>me</hello></response>'
    """
    async def get(self, request, entry):
        return {'hello': entry}


api.add_resource(Hello, '/<entry:string>')

if __name__ == '__main__':
    app.run(debug=True)



from sanic import Sanic, Blueprint
from sanic_restful import Api, Resource


class TestApi(Resource):
    async def get(self, request):
        return ''


class TestBlueprint:
    def test_register_blueprint(self):
        app = Sanic(__name__)
        bp = Blueprint('api', url_prefix='/api')
        api = Api(bp)
        api.add_resource(TestApi)
        app.blueprint(bp)

    def test_register_resources(self):
        bp = Blueprint('api', url_prefix='/api')
        api = Api()
        api.add_resource(TestApi)
        api.init_app(bp)
        app = Sanic(__name__)
        app.blueprint(bp)
